/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SWO_DEBUG	FALSE
#define FE_SWO_VERBOSE	FALSE

#include "operator/operator.pmh"

using namespace fe;

//* TODO Soft-Bind - as though allowing multiple bindings per curve,
//* process multiple curve samples for each binding

//* TODO NaN check in driver, see std::numeric_limits<>::quiet_NaN()

void SurfaceWrapOp::initialize(void)
{
	//* NOTE inherits attributes from SurfaceBindOp

	catalogRemove("FragmentAttr");
	catalogRemove("DriverGroup");
	catalogRemove("Bindings");
	catalogRemove("Bind Primitives");
	catalogRemove("Bind Fragments");
	catalogRemove("Conform");
	catalogRemove("LimitDistance");
	catalogRemove("MaxDistance");
	catalogRemove("PartitionDriver");
	catalogRemove("InputPartitionAttr");
	catalogRemove("DriverPartitionAttr");
	catalogRemove("PivotPoint");

	catalog<String>("BindPrefix","hint")=
			"Prefix of attribute names on input surface describing"
			" where each point on the input surface is attached to"
			" the associated driver.";

	catalog<bool>("DiscardBindings")=true;
	catalog<String>("DiscardBindings","hint")=
			"Remove used binding attributes from the output.";

	catalog<Real>("Distance Bias")=10.0;
	catalog<Real>("Distance Bias","min")=0.0;
	catalog<Real>("Distance Bias","high")=20.0;
	catalog<Real>("Distance Bias","max")=100.0;
	catalog<bool>("Distance Bias","joined")=true;
	catalog<String>("Distance Bias","hint")=
			"Dominance of closer faces towards weighting";

	catalog<Real>("Distance Scale")=1.0;
	catalog<Real>("Distance Scale","min")=0.0;
	catalog<Real>("Distance Scale","high")=2.0;
	catalog<Real>("Distance Scale","max")=100.0;
	catalog<bool>("Distance Scale","joined")=true;
	catalog<String>("Distance Scale","hint")=
			"Distance gain for weighting";

	catalog<bool>("singleTransform")=false;
	catalog<String>("singleTransform","label")="Single Transform";
	catalog<String>("singleTransform","hint")=
			"Use trasform from first binding only."
			"  Extra bindings simply translate in that space.";

	catalog<bool>("rotateNormals")=false;
	catalog<String>("rotateNormals","label")="Rotate Normals";
	catalog<String>("rotateNormals","hint")=
			"Apply transforms to normals in addition to positions.";

	catalog<bool>("checkDeformed")=false;
	catalog<String>("checkDeformed","label")="Check Deformed Driver";
	catalog<bool>("checkDeformed","joined")=true;
	catalog<String>("checkDeformed","hint")=
			"Verify that the deformed surface is like the driver surface."
			"  These checks take time, so this option should probably"
			" only be used temporarily.";

	catalog<I32>("checkLimit")=1;
	catalog<I32>("checkLimit","min")=0;
	catalog<I32>("checkLimit","high")=10;
	catalog<I32>("checkLimit","max")=100;
	catalog<String>("checkLimit","enabler")="checkDeformed";
	catalog<String>("checkLimit","label")="Message Limit";
	catalog<bool>("checkLimit","joined")=true;
	catalog<String>("checkLimit","hint")=
			"Stop reporting mismatches after the given number of errors.";

	catalog<Real>("scaleLimit")=2;
	catalog<Real>("scaleLimit","min")=1;
	catalog<Real>("scaleLimit","high")=10;
	catalog<String>("scaleLimit","enabler")="checkDeformed";
	catalog<String>("scaleLimit","label")="Scale Limit";
	catalog<String>("scaleLimit","hint")=
			"Threshold for rescale warnings on deformed driver.";

	catalog<bool>("Debug Bind")=false;
	catalog<bool>("Debug Bind","joined")=true;
	catalog<String>("Debug Bind","hint")=
			"Draws lines over the output showing the attachments."
			"  This feedback may only be visible is this node is"
			" \"selected\" and/or \"displayed\".";

	catalog<I32>("Debug Index")=-1;
	catalog<I32>("Debug Index","min")=-1;
	catalog<I32>("Debug Index","high")=1000;
	catalog<I32>("Debug Index","max")=1000000;
	catalog<String>("Debug Index","enabler")="Debug Bind";
	catalog<String>("Debug Index","hint")=
			"Index of element to examine (-1 to see them all).";

	catalog<String>("Input Surface","label")="Bound Input Surface";

	catalog< sp<Component> >("Deformed Surface");
	catalog<String>("Deformed Surface","label")="Deformed Driver Surface";

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";

	//* check problems
	catalog<bool>("Output Surface","recycle")=true;

	m_spDebugGroup=new DrawMode();
	m_spDebugGroup->setGroup("debug");
}

void SurfaceWrapOp::handle(Record& a_rSignal)
{
#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::handle\n");
#endif

	//* output curves is copy of input surface
	if(!accessOutput(m_spOutputAccessible,a_rSignal)) return;

	if(catalog<BWORD>("Debug Bind"))
	{
		if(!accessGuide(m_spDrawDebug,a_rSignal)) return;
	}
	else
	{
		m_spDrawDebug=NULL;
	}

	if(!access(m_spInputAccessible,"Input Surface")) return;

	sp<SurfaceAccessibleI> spDriverAccessible;
	if(!access(spDriverAccessible,"Driver Surface")) return;

	sp<SurfaceAccessorI> spDriverVertices;
	if(!access(spDriverVertices,spDriverAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spDeformedVertices;
	if(!access(spDeformedVertices,"Deformed Surface",
			e_primitive,e_vertices)) return;

	const String driverSignature=generateSignature(spDriverAccessible);
	const String inputSignature=retrieveSignature(m_spInputAccessible);
	if(driverSignature!=inputSignature)
	{
		String message;
		message.sPrintf("Reference driver surface has a different signature"
				" than when bound (\"%s\" vs \"%s\");",
				driverSignature.c_str(),inputSignature.c_str());
		catalog<String>("warning")+=message;
	}

	String driverGroup;
	String fragmentAttribute;
	String covertKey;
	retrieveParameters(m_spInputAccessible,driverGroup,fragmentAttribute,
			m_bindCount,m_bindPrimitives,m_fragmented,covertKey);

	m_spBindingAccessible=m_spInputAccessible;
	if(!covertKey.empty())
	{
		sp<SurfaceAccessibleCatalog> spSurfaceAccessibleCatalog=
				registry()->create("*.SurfaceAccessibleCatalog");
		if(spSurfaceAccessibleCatalog.isValid())
		{
			spSurfaceAccessibleCatalog->setCatalog(
					registry()->master()->catalog());
			spSurfaceAccessibleCatalog->setKey(covertKey);
			m_spBindingAccessible=spSurfaceAccessibleCatalog;
		}
	}

	const BWORD checkDeformed=catalog<bool>("checkDeformed");
	const U32 driverPrimitiveCount=spDriverVertices->count();
	const U32 deformedPrimitiveCount=spDeformedVertices->count();
	if(deformedPrimitiveCount!=driverPrimitiveCount)
	{
		String message;
		message.sPrintf("Deformed driver surface has a different number"
				" of primitives than reference (%d vs %d);",
				deformedPrimitiveCount,driverPrimitiveCount);
		catalog<String>(checkDeformed? "error": "warning")+=message;
		if(checkDeformed)
		{
			return;
		}
	}
	else if(checkDeformed)
	{
		sp<SurfaceAccessorI> spDriverPoints;
		if(!access(spDriverPoints,spDriverAccessible,
				e_point,e_position)) return;

		sp<SurfaceAccessorI> spDeformedPoints;
		if(!access(spDeformedPoints,"Deformed Surface",
				e_point,e_position)) return;

		const I32 checkLimit=catalog<I32>("checkLimit");
		const Real scaleLimit=catalog<Real>("scaleLimit");
		const Real invScaleLimit=scaleLimit>0.0? 1.0/scaleLimit: 0.0;
		U32 warnings=0;
		U32 mismatches=0;
		for(U32 primitiveIndex=0;
				primitiveIndex<driverPrimitiveCount && mismatches<checkLimit;
				primitiveIndex++)
		{
			const U32 driverSubCount=
					spDriverVertices->subCount(primitiveIndex);
			const U32 deformedSubCount=
					spDeformedVertices->subCount(primitiveIndex);
			if(driverSubCount!=deformedSubCount)
			{
				String message;
				message.sPrintf("  primitive %d has different"
						" number of vertices (%d vs %d);\n",primitiveIndex,
						deformedPrimitiveCount,driverPrimitiveCount);
				catalog<String>("error")+=message;
				mismatches++;
			}
			else
			{
				BWORD broken=FALSE;
				SpatialVector lastDriverPoint;
				SpatialVector lastDeformedPoint;
				for(U32 subIndex=0;
						subIndex<driverSubCount &&
						warnings+mismatches<checkLimit;
						subIndex++)
				{
					const U32 driverPointIndex=spDriverVertices->integer(
							primitiveIndex,subIndex);
					const U32 deformedPointIndex=spDeformedVertices->integer(
							primitiveIndex,subIndex);
					const SpatialVector driverPoint=
							spDriverPoints->spatialVector(driverPointIndex);
					const SpatialVector deformedPoint=
							spDeformedPoints->spatialVector(deformedPointIndex);

					if(FE_INVALID_SCALAR(driverPoint[0]) ||
							FE_INVALID_SCALAR(driverPoint[1]) ||
							FE_INVALID_SCALAR(driverPoint[2]))
					{
						String message;
						message.sPrintf(
								"  driver primitive %d vertex %d"
								" has corrupt data;\n",
								primitiveIndex,subIndex);
						catalog<String>("error")+=message;
						mismatches++;
					}
					else if(FE_INVALID_SCALAR(deformedPoint[0]) ||
							FE_INVALID_SCALAR(deformedPoint[1]) ||
							FE_INVALID_SCALAR(deformedPoint[2]))
					{
						String message;
						message.sPrintf(
								"  deformed primitive %d vertex %d"
								" has corrupt data;\n",
								primitiveIndex,subIndex);
						catalog<String>("error")+=message;
						mismatches++;
					}
					else if(driverPointIndex!=deformedPointIndex)
					{
						String message;
						message.sPrintf("  primitive %d has mismatched"
								" vertex %d (%d vs %d);\n",
								primitiveIndex,subIndex,
								deformedPointIndex,driverPointIndex);
						catalog<String>("error")+=message;
						mismatches++;
					}
					else if(subIndex && !broken)
					{
						const Real driverLength=
								magnitude(driverPoint-lastDriverPoint);
						const Real deformedLength=
								magnitude(deformedPoint-lastDeformedPoint);

						if(driverLength<1e-6)
						{
							String message;
							message.sPrintf(
									"  driver primitive %d segment %d"
									" has near zero length;\n",
									primitiveIndex,subIndex-1);
							catalog<String>("error")+=message;
							mismatches++;
						}
						else if(deformedLength<1e-6)
						{
							String message;
							message.sPrintf(
									"  deformed primitive %d segment %d"
									" has near zero length;\n",
									primitiveIndex,subIndex-1);
							catalog<String>("error")+=message;
							mismatches++;
						}
						else
						{
							const Real segmentScale=deformedLength/driverLength;
							if(segmentScale<invScaleLimit ||
									segmentScale>scaleLimit)
							{
								String message;
								message.sPrintf(
										"  deformed primitive %d segment %d"
										" has scaled by %.6G\n",
										primitiveIndex,subIndex-1,
										deformedLength/driverLength);
								catalog<String>("warning")+=message;
								warnings++;
							}
						}
					}
					lastDriverPoint=driverPoint;
					lastDeformedPoint=deformedPoint;
				}
			}
		}
		if(mismatches)
		{
			String message;
			message.sPrintf("Deformed driver has %d%s mismatches:\n",
					mismatches,mismatches>=checkLimit? "+": "");
			catalog<String>("error")=message+catalog<String>("error");
			return;
		}
	}

	//* original driver
	if(!access(m_spDriver,"Driver Surface",driverGroup)) return;

	m_spDriver->setRefinement(0);
	m_spDriver->prepareForSample();
	if(m_spDriver->radius()<1e-6)
	{
		catalog<String>("error")="Reference Driver Surface appears empty"
				" (zero radius bounding sphere);";
		return;
	}

	//* deformed driver
	if(!access(m_spDeformed,"Deformed Surface",driverGroup)) return;

	m_spDeformed->setRefinement(0);
	m_spDeformed->prepareForSample();
	if(m_spDeformed->radius()<1e-6)
	{
		catalog<String>("error")="Deformed Driver Surface appears empty"
				" (zero radius bounding sphere);";
		return;
	}

	m_element=m_bindPrimitives? e_primitive: e_point;
	m_attribute=m_bindPrimitives? e_vertices: e_position;

	sp<SurfaceAccessorI> spOutputElements;
	if(!access(spOutputElements,m_spOutputAccessible,
			m_element,m_attribute)) return;

	if(catalog<bool>("rotateNormals"))
	{
		sp<SurfaceAccessorI> spInputNormals;
		sp<SurfaceAccessorI> spOutputNormals;

		if(!access(spInputNormals,m_spInputAccessible,e_point,e_normal) ||
				!access(spOutputNormals,m_spOutputAccessible,e_point,e_normal))
		{
			catalog<String>("warning")+="can not rotate missing normals;";
		}
		else
		{
			//* flush normals
			const U32 normalCount=spInputNormals->count();
			for(U32 normalIndex=0;normalIndex<normalCount;normalIndex++)
			{
				spOutputNormals->set(normalIndex,
						spInputNormals->spatialVector(normalIndex));
			}
		}
	}

	const U32 elementCount=spOutputElements->count();

	if(m_spDrawDebug.isValid())
	{
		m_spDrawDebug->pushDrawMode(m_spDebugGroup);
	}

	U32 runCount=elementCount;
	m_spFragmentAccessor=NULL;
	if(m_fragmented)
	{
		if(access(m_spFragmentAccessor,m_spInputAccessible,
				m_element,m_attribute))
		{
			m_spFragmentAccessor->fragmentWith(
					SurfaceAccessibleI::Element(m_element),fragmentAttribute);
			runCount=m_spFragmentAccessor->fragmentCount();
			if(runCount<1)	//* was (runCount<2), but there can be one fragment
			{
				catalog<String>("warning")+=
						"fragmentation failed to divide surface -> reverting";
				m_fragmented=FALSE;
				runCount=elementCount;
			}
		}
		else
		{
			catalog<String>("warning")+=
					"fragmentation access failed -> ignoring";
			m_fragmented=FALSE;
			runCount=elementCount;
		}
	}

	//* TODO fragment-aware ranges
	if(m_fragmented)
	{
		//* serial
		sp<SurfaceRange> spFullRange=sp<SurfaceRange>(new SurfaceRange());
		spFullRange->nonAtomic().add(0,runCount-1);
		run(-1,spFullRange);
	}
	else
	{
		setAtomicRange(m_spOutputAccessible,
				m_bindPrimitives? e_pointsOfPrimitives: e_pointsOnly);

		OperatorThreaded::handle(a_rSignal);
	}

	if(m_spDrawDebug.isValid())
	{
		m_spDrawDebug->popDrawMode();
	}

	if(catalog<bool>("DiscardBindings"))
	{
		discardBindings(m_spOutputAccessible,m_bindCount,m_bindPrimitives);
		discardSignature(m_spOutputAccessible);
		discardParameters(m_spOutputAccessible);
	}

#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::handle done\n");
#endif
}

void SurfaceWrapOp::run(I32 a_id,sp<SurfaceRange> a_spRange)
{
	const BWORD singleTransform=catalog<bool>("singleTransform");
	const BWORD rotateNormals=catalog<bool>("rotateNormals");
	const Real distanceBias=catalog<Real>("Distance Bias");
	const Real distanceScale=catalog<Real>("Distance Scale");
	const I32 debugIndex=catalog<I32>("Debug Index");

	const BWORD cacheOutput=FALSE;

#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::run running thread %d %s\n",
			a_id,a_spRange->brief().c_str());
#endif

	sp<SurfaceAccessorI> spInputElements;
	if(!access(spInputElements,m_spInputAccessible,
			m_element,m_attribute)) return;

	sp<SurfaceAccessorI> spOutputElements;
	if(cacheOutput)
	{
		sp<SurfaceAccessorI> spOutputElementsDirect;
		if(!access(spOutputElementsDirect,m_spOutputAccessible,
				m_element,m_attribute)) return;

		sp<SurfaceAccessorCached> spAccessorCached=
				registry()->create("*.SurfaceAccessorCached");
		FEASSERT(spAccessorCached.isValid());
		spAccessorCached->setSurfaceAccessible(m_spOutputAccessible);
		spAccessorCached->setAccessorChain(spOutputElementsDirect);

		spOutputElements=spAccessorCached;
	}
	else
	{
		if(!access(spOutputElements,m_spOutputAccessible,
				m_element,m_attribute)) return;
	}

	sp<SurfaceAccessorI> spInputNormals;
	sp<SurfaceAccessorI> spOutputNormals;
	if(rotateNormals)
	{
		access(spInputNormals,m_spInputAccessible,
				e_point,e_normal,e_quiet);
		access(spOutputNormals,m_spOutputAccessible,
				e_point,e_normal,e_quiet);
	}

	std::vector< sp<SurfaceAccessorI> > inputFaces;
	std::vector< sp<SurfaceAccessorI> > inputBarys;
	const I32 bindingsAccessed=accessBindings(m_spBindingAccessible,
			inputFaces,inputBarys,m_bindCount,m_bindPrimitives,FALSE);

	if(bindingsAccessed<1)
	{
		return;
	}

	const BWORD singular=
			(m_fragmented && m_element==SurfaceAccessibleI::e_point);

	sp<SurfaceAccessorI::FilterI> spFilter;

	for(SurfaceRange::Iterator it=a_spRange->begin();!it.atEnd();it.step())
	{
		const U32 index=it.value();

		I32 filterCount=1;
		U32 tableCount=0;
		if(m_fragmented)
		{
			const String fragment=m_spFragmentAccessor->fragment(index);
			if(!m_spFragmentAccessor->filterWith(fragment,spFilter) ||
					spFilter.isNull())
			{
				continue;
			}
			filterCount=spFilter->filterCount();
			if(filterCount<1)
			{
				continue;
			}
			if(singular)
			{
				tableCount+=filterCount;
			}
			else
			{
				for(I32 filterIndex=0;filterIndex<filterCount;filterIndex++)
				{
					const I32 filtered=spFilter->filter(filterIndex);
					FEASSERT(filtered>=0);
					tableCount+=spInputElements->subCount(filtered);
				}
			}
		}
		else
		{
			tableCount=spInputElements->subCount(index);
		}

		if(tableCount<1)
		{
			continue;
		}

		SpatialVector point[tableCount];
		SpatialVector normal[tableCount];
		SpatialVector resultPoint[tableCount];
		SpatialVector resultNormal[tableCount];
		Real totalWeight[tableCount];

		U32 tableIndex=0;
		for(U32 filterIndex=0;filterIndex<filterCount;filterIndex++)
		{
			const I32 filtered=
					m_fragmented? spFilter->filter(filterIndex): index;
			FEASSERT(filtered>=0);
			const U32 subCount=singular? 1: spInputElements->subCount(filtered);
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				point[tableIndex]=
						spInputElements->spatialVector(filtered,subIndex);
				if(spInputNormals.isValid())
				{
					const I32 pointIndex=m_bindPrimitives?
							spInputElements->integer(filtered,subIndex):
							filtered;
					normal[tableIndex]=
							spInputNormals->spatialVector(pointIndex);
				}
				set(resultPoint[tableIndex]);
				set(resultNormal[tableIndex]);
				totalWeight[tableIndex]=0.0;
				tableIndex++;
			}
		}

		const I32 keyIndex=m_fragmented? spFilter->filter(0): index;
		FEASSERT(keyIndex>=0);

		SpatialTransform coreConversion;

		for(U32 binding=0;binding<bindingsAccessed;binding++)
		{
			const I32 face=inputFaces[binding]->integer(keyIndex);
			if(face<0)
			{
				continue;
			}

			SpatialTransform conversion;
			SpatialVector translateDriver;
			SpatialVector translateDeformed;

			SpatialBary barycenter(0.33,0.33);
			if(!binding)
			{
				barycenter=inputBarys[binding]->spatialVector(keyIndex);
			}

			if(!binding || !singleTransform)
			{
				const SpatialTransform xformDriver=
						m_spDriver->sample(face,barycenter);
				const SpatialTransform xformDeformed=
						m_spDeformed->sample(face,barycenter);

				SpatialTransform invDriver;
				invert(invDriver,xformDriver);
				conversion=invDriver*xformDeformed;
				coreConversion=conversion;

				translateDriver=xformDriver.translation();
				translateDeformed=xformDeformed.translation();
			}
			else
			{
				//* TODO sample translation only
				translateDriver=
						m_spDriver->sample(face,barycenter).translation();
				translateDeformed=
						m_spDeformed->sample(face,barycenter).translation();

				SpatialVector refDeformed;
				transformVector(coreConversion,translateDriver,refDeformed);

				conversion=coreConversion;
				conversion.translation()+=translateDeformed-refDeformed;
			}

			const Real distance=magnitude(translateDriver-point[0]);
			const Real weight=
					1.0/(1.0+pow(distanceScale*distance,distanceBias));

			for(tableIndex=0;tableIndex<tableCount;tableIndex++)
			{
				totalWeight[tableIndex]+=weight;

				SpatialVector transformedPoint;
				transformVector(conversion,point[tableIndex],transformedPoint);
				resultPoint[tableIndex]+=weight*transformedPoint;

				if(rotateNormals)
				{
					SpatialVector transformedNormal;
					rotateVector(conversion,normal[tableIndex],
							transformedNormal);
					resultNormal[tableIndex]+=weight*transformedNormal;
				}

#if FE_SWO_VERBOSE
				feLog("SurfaceWrapOp::handle element %d bind %d/%d"
						" point %d/%d face %d bary %s weight %.6G"
						" from %s to %s\n",
						index,binding,bindingsAccessed,tableIndex,tableCount,
						face,c_print(barycenter),weight,
						c_print(point[tableIndex]),c_print(transformedPoint));
				feLog("conversion\n%s\n",c_print(conversion));
				feLog("xformDriver\n%s\n",c_print(xformDriver));
				feLog("xformDeformed\n%s\n",c_print(xformDeformed));
#endif
			}

			if(m_spDrawDebug.isValid() && (debugIndex<0 || debugIndex==index))
			{
				SpatialVector transformedPoint;
				transformVector(conversion,point[0],transformedPoint);

				const Color yellow(1.0,1.0,0.0,1.0);
				SpatialVector line[2];

				line[0]=transformedPoint;
				line[1]=translateDeformed;

				SAFEGUARD;
				m_spDrawDebug->drawLines(line,NULL,2,DrawI::e_strip,
						false,&yellow);
			}
		}

		tableIndex=0;
		for(U32 filterIndex=0;filterIndex<filterCount;filterIndex++)
		{
			const I32 filtered=
					m_fragmented? spFilter->filter(filterIndex): index;
			FEASSERT(filtered>=0);
			const U32 subCount=singular? 1: spInputElements->subCount(filtered);
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				if(totalWeight[tableIndex]>0.0)
				{
					spOutputElements->set(filtered,subIndex,
							resultPoint[tableIndex]/totalWeight[tableIndex]);
					if(rotateNormals && spOutputNormals.isValid())
					{
						const I32 pointIndex=m_bindPrimitives?
								spOutputElements->integer(filtered,subIndex):
								filtered;

						spOutputNormals->set(pointIndex,
								unitSafe(resultNormal[tableIndex]));
					}
				}

				tableIndex++;
			}
		}
	}

#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::run done running thread %d\n",a_id);
#endif
}
